"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var typescript_1 = __importDefault(require("typescript"));
var ISerializeSymbol_1 = require("./ISerializeSymbol");
function generateDocumentation(fileNames, options) {
    if (options === void 0) { options = {
        module: typescript_1.default.ModuleKind.CommonJS,
        target: typescript_1.default.ScriptTarget.ES2015
    }; }
    var program = typescript_1.default.createProgram(fileNames, options);
    var checker = program.getTypeChecker();
    var output = [];
    program.getSourceFiles()
        .forEach(function (sourceFile) {
        if (!sourceFile.isDeclarationFile) {
            typescript_1.default.forEachChild(sourceFile, visit);
        }
    });
    return output;
    function visit(node) {
        if (!isNodeExported(node)) {
            return;
        }
        if (node.kind === typescript_1.default.SyntaxKind.ClassDeclaration) {
            var currentNode = node;
            if (currentNode.name === undefined) {
                return;
            }
            var symbol = checker.getSymbolAtLocation(currentNode.name);
            if (symbol === undefined) {
                return;
            }
            output.push(serializeClass(symbol));
            return;
        }
        if (node.kind === typescript_1.default.SyntaxKind.InterfaceDeclaration) {
            var currentNode = node;
            var symbol = checker.getSymbolAtLocation(currentNode.name);
            if (symbol === undefined) {
                return;
            }
            output.push(serializeInterface(symbol));
            return;
        }
        if (node.kind === typescript_1.default.SyntaxKind.ModuleDeclaration) {
            typescript_1.default.forEachChild(node, visit);
        }
        if (node.kind === typescript_1.default.SyntaxKind.EnumDeclaration) {
            var currentNode = node;
            var symbol = checker.getSymbolAtLocation(currentNode.name);
            if (symbol === undefined) {
                return;
            }
            output.push(serializeEnum(symbol));
            return;
        }
    }
    function serializeSymbol(symbol) {
        return {
            name: symbol.getName(),
            type: checker.typeToString(checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration))
        };
    }
    function serializeMember(symbol) {
        var result = {
            keyword: getMemberKeyword(symbol),
            modifierType: getMemberModifierType(symbol),
            name: symbol.getName(),
            parameters: [],
            returnType: getMemberReturnType(symbol),
            type: getMemberType(symbol)
        };
        if (result.type === ISerializeSymbol_1.MEMBER_TYPE.CONSTRUCTOR) {
            return;
        }
        if (result.type === ISerializeSymbol_1.MEMBER_TYPE.METHOD) {
            result.parameters = getParametersForFunction(checker.typeToString(checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration)));
        }
        return result;
    }
    function getMemberReturnType(symbol) {
        var returnType = getMemberType(symbol);
        if (returnType === ISerializeSymbol_1.MEMBER_TYPE.PROPERTY || returnType === ISerializeSymbol_1.MEMBER_TYPE.INDEX) {
            return checker.typeToString(checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration));
        }
        if (returnType === ISerializeSymbol_1.MEMBER_TYPE.METHOD) {
            return getReturnTypeOfFunction(checker.typeToString(checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration)));
        }
        if (returnType === ISerializeSymbol_1.MEMBER_TYPE.CONSTRUCTOR) {
            return '';
        }
        throw new Error('unable to determine return type');
    }
    function getParametersForFunction(typeString) {
        var parsedType = typeString.substring(1);
        parsedType = parsedType.substring(0, parsedType.indexOf(') => '));
        if (parsedType.length === 0) {
            return [];
        }
        var parameters = parsedType.split(', ');
        return parameters.map(function (parameter) {
            return {
                name: parameter.substring(0, parameter.indexOf(':')),
                type: parameter.substring(parameter.indexOf(':') + 2)
            };
        });
    }
    function getReturnTypeOfFunction(typeString) {
        return typeString.substring(typeString.lastIndexOf(' => ') + 4);
    }
    function getMemberType(symbol) {
        if (symbol.valueDeclaration !== undefined) {
            if (symbol.valueDeclaration.kind === typescript_1.default.SyntaxKind.PropertyDeclaration ||
                symbol.valueDeclaration.kind === typescript_1.default.SyntaxKind.PropertySignature ||
                symbol.valueDeclaration.kind === typescript_1.default.SyntaxKind.Parameter) {
                return ISerializeSymbol_1.MEMBER_TYPE.PROPERTY;
            }
            if (symbol.valueDeclaration.kind === typescript_1.default.SyntaxKind.MethodDeclaration ||
                symbol.valueDeclaration.kind === typescript_1.default.SyntaxKind.MethodSignature) {
                return ISerializeSymbol_1.MEMBER_TYPE.METHOD;
            }
            if (symbol.valueDeclaration.kind === typescript_1.default.SyntaxKind.Constructor) {
                return ISerializeSymbol_1.MEMBER_TYPE.CONSTRUCTOR;
            }
            if (symbol.valueDeclaration.kind === typescript_1.default.SyntaxKind.GetAccessor) {
                return ISerializeSymbol_1.MEMBER_TYPE.PROPERTY;
            }
            if (symbol.valueDeclaration.kind === typescript_1.default.SyntaxKind.SetAccessor) {
                return ISerializeSymbol_1.MEMBER_TYPE.PROPERTY;
            }
        }
        else if (symbol.declarations !== undefined && symbol.declarations.length > 0) {
            var kind_1;
            symbol.declarations.some(function (declaration) {
                if (declaration.kind === typescript_1.default.SyntaxKind.PropertyDeclaration ||
                    declaration.kind === typescript_1.default.SyntaxKind.PropertySignature ||
                    declaration.kind === typescript_1.default.SyntaxKind.Parameter) {
                    kind_1 = ISerializeSymbol_1.MEMBER_TYPE.PROPERTY;
                    return true;
                }
                if (declaration.kind === typescript_1.default.SyntaxKind.MethodDeclaration ||
                    declaration.kind === typescript_1.default.SyntaxKind.MethodSignature) {
                    kind_1 = ISerializeSymbol_1.MEMBER_TYPE.METHOD;
                    return true;
                }
                if (declaration.kind === typescript_1.default.SyntaxKind.Constructor) {
                    kind_1 = ISerializeSymbol_1.MEMBER_TYPE.CONSTRUCTOR;
                    return true;
                }
                if (declaration.kind === typescript_1.default.SyntaxKind.IndexSignature) {
                    kind_1 = ISerializeSymbol_1.MEMBER_TYPE.INDEX;
                    return true;
                }
                return false;
            });
            if (kind_1 === undefined) {
                throw new Error('unable to determine member type');
            }
            return kind_1;
        }
        throw new Error('unable to determine member type');
    }
    function getMemberModifierType(symbol) {
        var keyword = ISerializeSymbol_1.MODIFIER_TYPE.PUBLIC;
        if (symbol.valueDeclaration === undefined ||
            symbol.valueDeclaration.modifiers === undefined ||
            symbol.valueDeclaration.modifiers.length === 0) {
            return keyword;
        }
        symbol.valueDeclaration.modifiers.some(function (modifier) {
            if (modifier.kind === typescript_1.default.SyntaxKind.PrivateKeyword) {
                keyword = ISerializeSymbol_1.MODIFIER_TYPE.PRIVATE;
                return true;
            }
            if (modifier.kind === typescript_1.default.SyntaxKind.PublicKeyword) {
                keyword = ISerializeSymbol_1.MODIFIER_TYPE.PUBLIC;
                return true;
            }
            if (modifier.kind === typescript_1.default.SyntaxKind.ProtectedKeyword) {
                keyword = ISerializeSymbol_1.MODIFIER_TYPE.PROTECTED;
                return true;
            }
            return false;
        });
        return keyword;
    }
    function getMemberKeyword(symbol) {
        var keyword;
        if (symbol.valueDeclaration === undefined ||
            symbol.valueDeclaration.modifiers === undefined ||
            symbol.valueDeclaration.modifiers.length === 0) {
            return keyword;
        }
        symbol.valueDeclaration.modifiers.some(function (modifier) {
            if (modifier.kind === typescript_1.default.SyntaxKind.AbstractKeyword) {
                keyword = ISerializeSymbol_1.KEYWORD_TYPE.ABSTRACT;
                return true;
            }
            if (modifier.kind === typescript_1.default.SyntaxKind.StaticKeyword) {
                keyword = ISerializeSymbol_1.KEYWORD_TYPE.STATIC;
                return true;
            }
            return false;
        });
        return keyword;
    }
    function getExtendsClassName(heritageClause) {
        return heritageClause.types[0].expression.text;
    }
    function getInterfaceName(nodeObject) {
        return nodeObject.expression.text;
    }
    function getInterfaceNames(heritageClause) {
        return heritageClause.types.map(getInterfaceName);
    }
    function serializeInterface(symbol) {
        var serializedInterface = __assign({}, serializeSymbol(symbol), { members: [], structure: ISerializeSymbol_1.STRUCTURE.INTERFACE });
        if (symbol.members === undefined) {
            return serializedInterface;
        }
        symbol.members.forEach(function (memberName) {
            var serializedMember = serializeMember(memberName);
            if (serializedMember === undefined) {
                return;
            }
            serializedInterface.members.push(serializedMember);
        });
        return serializedInterface;
    }
    function serializeEnum(symbol) {
        var serializedEnum = __assign({}, serializeSymbol(symbol), { members: [], structure: ISerializeSymbol_1.STRUCTURE.ENUM });
        if (symbol.exports === undefined) {
            return serializedEnum;
        }
        symbol.exports.forEach(function (memberName) {
            serializedEnum.members.push(memberName.name);
        });
        return serializedEnum;
    }
    function serializeClass(symbol) {
        var constructorType = checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration);
        var details = __assign({}, serializeSymbol(symbol), { constructors: constructorType.getConstructSignatures()
                .map(serializeSignature), keyword: getMemberKeyword(symbol), members: [], structure: ISerializeSymbol_1.STRUCTURE.CLASS });
        if (symbol.members !== undefined) {
            symbol.members.forEach(function (memberName) {
                var serializedMember = serializeMember(memberName);
                if (serializedMember === undefined) {
                    return;
                }
                details.members.push(serializedMember);
            });
        }
        if (symbol.valueDeclaration.members !== undefined) {
            symbol.valueDeclaration.members.forEach(function (memberName) {
                var classSymbol = checker.getSymbolAtLocation(memberName.name);
                if (classSymbol === undefined) {
                    return;
                }
                var serializedMember = serializeMember(classSymbol);
                if (serializedMember === undefined) {
                    return;
                }
                if (!details.members.some(function (e) { return e.name === serializedMember.name; })) {
                    details.members.push(serializedMember);
                }
            });
        }
        var clauses = symbol.valueDeclaration.heritageClauses;
        if (clauses === undefined) {
            return details;
        }
        for (var _i = 0, clauses_1 = clauses; _i < clauses_1.length; _i++) {
            var heritageClause = clauses_1[_i];
            if (heritageClause.token === typescript_1.default.SyntaxKind.ExtendsKeyword) {
                details.extends = getExtendsClassName(heritageClause);
            }
            else if (heritageClause.token === typescript_1.default.SyntaxKind.ImplementsKeyword) {
                details.implements = getInterfaceNames(heritageClause);
            }
            else {
                throw new Error('unsupported heritage clause');
            }
        }
        return details;
    }
    function serializeSignature(signature) {
        return {
            parameters: signature.parameters.map(serializeSymbol),
            returnType: checker.typeToString(signature.getReturnType())
        };
    }
    function isNodeExported(node) {
        return (node.flags & typescript_1.default.ModifierFlags.Export) !== 0 ||
            node.parent.kind === typescript_1.default.SyntaxKind.SourceFile;
    }
}
exports.generateDocumentation = generateDocumentation;
